from __future__ import annotations

from typing import Dict, List, Type, Union, TYPE_CHECKING


if TYPE_CHECKING:
    from .. import Table
    from .comparison import Comparison

class Statement:
    def __init__(self, table:Table) -> None:
        self.table = table
        self.declarations = []
        self.preparations = []
    
    def where(self, comparisons:Comparison) -> Type[Statement]:
        self.preparations.extend(comparisons.preparations)
        comparisons.db_type = self.table.database.db_type
        self.declarations += ["WHERE", comparisons.render()]
        return self

    def inner_join(self, table:str, alias:str = ""):
        self.declarations.append(f"""INNER JOIN {table}{f" AS {alias}" if alias != "" else ""}""")
        return self

    def left_join(self, table:str, alias:str = ""):
        self.declarations.append(f"""LEFT JOIN {table}{f" AS {alias}" if alias != "" else ""}""")
        return self

    def right_join(self, table:str, alias:str = ""):
        self.declarations.append(f"""RIGHT JOIN {table}{f" AS {alias}" if alias != "" else ""}""")
        return self

    def full_join(self, table:str, alias:str = ""):
        self.declarations.append(f"""FULL JOIN {table}{f" AS {alias}" if alias != "" else ""}""")
        return self

    def on(self, comparisons:Comparison) -> Type[Statement]:
        self.preparations.extend(comparisons.preparations)
        comparisons.db_type = self.table.database.db_type
        self.declarations += ["ON", comparisons.render()]
        return self

    def render(self):
        return " ".join(self.declarations)

    def execute(self):
        cur = self.table.database.connection.cursor()
        ret_val = []
        if self.table.database.db_type == "sqlite":
            cur.execute(self.render(), tuple(self.preparations))
        elif self.table.database.db_type == "mysql":
            cur.execute(self.render(), tuple(self.preparations))
        
        ret_val = cur.fetchall()
        cur.close()
        self.table.database.connection.commit()
        return ret_val

class Select(Statement):
    def __init__(self, table:Table, modifier:str = "", parameters:List[Union[str, Statement, tuple]] = ["*"], alias:str = "", additional_tables:List[str] = "") -> None:
        self.table = table
        params_list = []
        for param in parameters:
            if type(param) == str:
                params_list.append(param)
            if type(param) == Statement:
                params_list.append("(" + param.render() + ")")
            if type(param) == tuple:
                params_list.append(" ".join(param))
                
        self.declarations = ["SELECT", modifier, ", ".join(params_list), "FROM", table.table_name, ", " + ", ".join(additional_tables) if len(additional_tables) != 0 else "", f"AS {alias}" if alias != "" else ""]
        self.preparations = []

class Delete(Statement):
    def __init__(self, table:Table) -> None:
        self.table = table
        self.declarations = ["DELETE FROM", table.table_name]
        self.preparations = []

class Insert(Statement):
    def __init__(self, table:Table, modifier:str = "", columns:List[str] = [], parameters:List[Union[str, Statement]] = [], alias = "") -> None:
        self.table = table
        self.preparations = []
        self.render_preps = []
        self.alias = alias
        for param in parameters:
            if type(param) == Statement:
                self.preparations.extend(param.preparations)
                self.render_preps.append("(" + param.render() + ")")
            else:
                self.preparations.append(param)
                if table.database.db_type == "sqlite":
                    self.render_preps.append("?")
                elif table.database.db_type == "mysql":
                    self.render_preps.append("%s")
        self.declarations = [
            "INSERT", "" if modifier == "" else f"OR {modifier}", 
            "INTO", table.table_name, 
            "" if alias == "" else f"AS {alias}", 
            f"""({", ".join(columns)})""" if columns != [] else "", 
            "VALUES", "(" + ", ".join(self.render_preps) + ")"
            ]

class Update(Statement):
    def __init__(self, table:Table, modifier:str = "", updates:Dict[str, str | int | bool | Statement] = {}, alias = "") -> None:
        self.table = table
        self.preparations = []
        self.render_preps = []
        self.alias = alias
        for key, val in updates.items():
            if type(val) in [str, int, bool]:
                self.preparations.append(val)
                if table.database.db_type == "sqlite":
                    self.render_preps.append(f"{key} = ?")
                elif table.database.db_type == "mysql":
                    self.render_preps.append(f"{key} = %s")
            if type(val) == Statement:
                self.preparations.extend(val.preparations)
                self.render_preps.append(f"{key} = (" + val.render() + ")")
        self.declarations = ["UPDATE", "" if modifier == "" else f"OR {modifier}", table.table_name, "" if alias == "" else f"AS {alias}", "SET", ", ".join(self.render_preps)]


from unittest import TestCase
from src.TwigORM.database import ColumnType, Database, Column, ForeignKey, SQLType
from src.TwigORM.database.statement.comparison import Equal
import unittest

unittest.TestLoader.sortTestMethodsUsing = None
dbname = "sqlite"

class Testing(TestCase):
    def test_1drop(self):
        db = Database(dbname, "test/testdb.db")
        db.drop("Post")
        db.drop("Users")
        db.close()

    def test_2createtable(self):
        db = Database(dbname, "test/testdb.db")
        db.add_table("Users",
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            Username = Column(ColumnType(SQLType.varchar, "30"), ["not null"]),
            Password = Column(ColumnType(SQLType.varchar, "64"), ["not null"])
        )
        db.close()

    def test_3insert(self):
        db = Database(dbname, "test/testdb.db")
        db.add_table("Users",
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            Username = Column(ColumnType(SQLType.varchar, "30"), ["not null"]),
            Password = Column(ColumnType(SQLType.varchar, "64"), ["not null"])
        )
        db.tables["Users"].insert(
            columns=["Username", "Password"],
            parameters=["William", "Lim123"]
        ).execute()
        db.close()

    def test_4select(self):
        db = Database(dbname, "test/testdb.db")
        db.add_table("Users",
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            Username = Column(ColumnType(SQLType.varchar, "30"), ["not null"]),
            Password = Column(ColumnType(SQLType.varchar, "64"), ["not null"])
        )
        self.assertEqual(db.tables["Users"].select().where(
            Equal("Username", ["William"]).Or(
                Equal("Username", ["Alex"])
            )
        ).execute(), [(1, 'William', 'Lim123')])
        db.close()

    def test_5update(self):
        db = Database(dbname, "test/testdb.db")
        db.add_table("Users",
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            Username = Column(ColumnType(SQLType.varchar, "30"), ["not null"]),
            Password = Column(ColumnType(SQLType.varchar, "64"), ["not null"])
        )
        db.tables["Users"].update(updates={
            "Username":"Alex"
        }).where(Equal("id", [1])).execute()
        db.close()

    def test_6delete(self):
        db = Database(dbname, "test/testdb.db")
        db.add_table("Users",
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            Username = Column(ColumnType(SQLType.varchar, "30"), ["not null"]),
            Password = Column(ColumnType(SQLType.varchar, "64"), ["not null"])
        )
        db.tables["Users"].delete().where(Equal("id", [2])).execute()
        db.close()

    def test_7inner_join(self):
        db = Database(dbname, "test/testdb.db")
        db.add_table("Users",
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            Username = Column(ColumnType(SQLType.varchar, "30"), ["not null"]),
            Password = Column(ColumnType(SQLType.varchar, "64"), ["not null"])
        )
        db.add_table("Post",
            [
                ForeignKey(["UserID"], "Users", ["id"])
            ],
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            UserID = Column(ColumnType(SQLType.integer), ["not null"]),
            title = Column(ColumnType(SQLType.varchar, "100"), ["not null"]),
            body = Column(ColumnType(SQLType.text), ["not null"])
        )
        db.tables["Post"].drop()
        db.add_table("Post",
            [
                ForeignKey(["UserID"], "Users", ["id"])
            ],
            id = Column(ColumnType(SQLType.integer), ["primary key"]),
            UserID = Column(ColumnType(SQLType.integer), ["not null"]),
            title = Column(ColumnType(SQLType.varchar, "100"), ["not null"]),
            body = Column(ColumnType(SQLType.text), ["not null"])
        )
        db.tables["Post"].insert(columns=["UserID", "title", "body"], parameters=[1, "This is a test Title.", "This is a test body."]).execute()
        self.assertEqual(
            db.tables["Users"].select(parameters=["Users.Username", "Post.title", "Post.body"]).inner_join("Post").on(Equal("Users.id", "Post.UserID")).execute(),
            [('Alex', 'This is a test Title.', 'This is a test body.')]
        )
        
        db.close()
    


from __future__ import annotations

import mysql.connector
import sqlite3
from typing import Dict, List
from .statement import Select, Statement, Insert, Update, Delete
from . import statement

class SQLType:
    Char = "char"
    varchar = "varchar"
    binary = "binary"
    varbinary = "varbinary"
    tinyblob = "tinyblob"
    tinytext = "tinytext"
    text = "text"
    blob = "blob"
    mediumtext = "mediumtext"
    mediumblob = "mediumblob"
    longtext = "longtext"
    longblob = "longblob"
    enum = "enum"
    set = "set"
    bit = "bit"
    tinyint = "tinyint"
    bool = "bool"
    boolean = "bool"
    smallint = "smallint"
    mediumint = "mediumint"
    Int = "int"
    integer = "integer"
    bigint = "bigint"
    float = "float"
    double = "double"
    double_precision = "double precision"
    decimal = "decimal"
    dec = "dec"
    date = "date"
    datetime = "datetime"
    timestamp = "timestamp"
    time = "time"
    year = "year"
    

class ColumnType:
    def __init__(self, sqltype:SQLType | str, *args:str) -> None:
        self.db_type = "sqlite"
        self.type = sqltype
        self.args = args
        self.db_type = "sqlite"

    def render(self):
        # Change column parameter based on db_type
        return f"""{self.type}{f"({', '.join(self.args)})" if len(self.args) != 0 else ""}"""


class Column:
    def __init__(self, type:ColumnType, constraints:List[str]) -> None:
        self.name = ""
        self.constraints = constraints
        self.db_type = "sqlite"
        self.type = type
    
    def __repr__(self) -> str:
        self.type.db_type = self.db_type
        return f"""{self.name} {self.type.render()} {" ".join(self.constraints)}"""

class ForeignKey:
    def __init__(self, columns:List[str], table:str, constraint_columns: List[str]) -> None:
        self.columns = columns
        self.constraints = constraint_columns
        self.db_type = "sqlite"
        self.table = table
        self.constraint_columns = constraint_columns

    def __repr__(self) -> str:
        return f"""FOREIGN KEY ({", ".join(self.columns)}) REFERENCES {self.table}({", ".join(self.constraint_columns)})"""

class Table:
    def __init__(self, database:Database, table_name:str, columns:Dict[str, Column], extra_constraints:List[ForeignKey]) -> None:
        self.database = database
        self.table_name = table_name
        self.columns = columns
        columns_str = ""
        for cn, (_, col) in enumerate(columns.items()):
            col.db_type = database.db_type
            if cn == len(columns)-1:
                columns_str += f"{col}"
            else:
                columns_str += f"{col}, "
        createSQL = f"""CREATE TABLE IF NOT EXISTS {table_name}({columns_str}{", " + ", ".join([f"{x}" for x in extra_constraints]) if len(extra_constraints) != 0 else ""});"""
        #print(createSQL)
        curs = database.connection.cursor()
        curs.execute(createSQL)
        #curs.close()
        database.connection.commit()
        
    
    def select(self, modifier = "", parameters: List[str | Statement | tuple] = ["*"], alias:str = "", additional_tables:List[str] = "") -> Select:
        return Select(self, modifier, parameters, alias, additional_tables)

    def insert(self, modifier = "", columns:List[str] = [], parameters: List[str | Statement | tuple] = [], alias = "") -> Select:
        return Insert(self, modifier, columns, parameters, alias)

    def update(self, modifier:str = "", updates:Dict[str, str | int | bool | Statement] = {}, alias = "") -> Select:
        return Update(self, modifier, updates, alias)
    
    def delete(self):
        return Delete(self)

    def drop(self):
        sql = f"DROP TABLE IF EXISTS {self.table_name}"
        curs = self.database.connection.cursor()
        if self.database.db_type == "sqlite":
            curs.execute(sql)
        elif self.database.db_type == "mysql":
            curs.execute(sql)
        #curs.close()
        self.database.connection.commit()


class Database:
    def __init__(self, db_type:str, db_address:str, user = "", password = "", database_name = "") -> None:
        if db_type.lower() == "sqlite":
            self.connection = sqlite3.connect(db_address, timeout=10)
            
        elif db_type.lower() == "mysql":
            params = {}
            if user != "":
                params['user'] = user
            if password != "":
                params['password'] = password
            if database_name != "":
                params['database'] = database_name

            params['host'] = db_address
            
            self.connection = mysql.connector.connect(**params)
            
        self.db_type = db_type
        self.tables:Dict[str, Table] = {}

    def create_database(self, database_name:str):
        if self.db_type == "sqlite":
            print("Error: SQLite3 does not have a CREATE DATABASE functionality.")
            return
        curs = self.connection.cursor()
        sql = f"CREATE DATABASE IF NOT EXISTS {database_name}"
        curs.execute(sql)
        #curs.close()
        self.connection.commit()

    def add_table(self, table_name:str, extra_constraints:List[ForeignKey] = [], **columns:Column):
        for key in columns.keys():
            columns[key].name = key
        self.tables[table_name] = Table(self, table_name, columns, extra_constraints)
        return self

    def drop(self, table:str):
        sql = f"DROP TABLE IF EXISTS {table};"
        curs = self.connection.cursor()
        if self.db_type == "sqlite":
            curs.execute(sql)
        elif self.db_type == "mysql":
            curs.execute(sql)
        #curs.close()
        self.connection.commit()

    def close(self):
        self.connection.close()


from __future__ import annotations

from typing import Union, Tuple, Type, List, TYPE_CHECKING

from . import Statement

class Comparison:
    #CLASS IS DESIGNED TO BE INHERITED
    def __init__(self, lhs:Union[str, int, float, Comparison, Statement], rhs:Union[str, int, float, Comparison, Statement]) -> None:
        self.operator = ""
        self.preparations = []
        if type(lhs) == Statement:
            self.lhs = f"({lhs.render()})"
        else:
            self.lhs = lhs

        if type(rhs) == Statement:
            self.rhs = f"({rhs.render()})"
        elif type(rhs) == list:
            self.rhs = "?"
            self.db_type = "sqlite"
            self.preparations.append(rhs[0])
        else:
            self.rhs = rhs

        self.additions:List[Union[str, int, float, Comparison]] = []
    
    def Or(self, comparison:Type[Comparison] | Tuple[Comparison] | List[Comparison]):
        self.additions.append("OR")
        if issubclass(type(comparison), Comparison):
            self.preparations.extend(comparison.preparations)
            self.additions.append(comparison)
        elif type(comparison) == tuple or type(comparison) == list:
            self.preparations.extend(comparison[0].preparations)
            self.additions.extend(["(", comparison[0], ")"])
        return self

    def And(self, comparison:Type[Comparison] | Tuple[Comparison] | List[Comparison]):
        self.additions.append("AND")
        if issubclass(type(comparison), Comparison):
            self.preparations.extend(comparison.preparations)
            self.additions.append(comparison)
        elif type(comparison) == tuple or type(comparison) == list:
            self.preparations.extend(comparison[0].preparations)
            self.additions.extend(["(", comparison[0], ")"])
        return self

    def render(self) -> str:
        compiled_additions = ""
        
        for an, addition in enumerate(self.additions):
            if issubclass(type(addition), Comparison):
                if self.db_type == "mysql":
                    addition.db_type = self.db_type
                compiled_additions += f"{addition.render()}"
            else:
                compiled_additions += f"{addition}"
            if an != len(self.additions) - 1:
                compiled_additions += " "
                
        if self.rhs == "?":

            if self.db_type == "mysql":
                self.rhs = "%s"
            elif self.db_type == "sqlite":
                self.rhs = "?"
        
        return f"{self.lhs} {self.operator} {self.rhs} " + compiled_additions

class Equal(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = "="

class NotEqual(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = "!="

class LessThan(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = "<"

class GreaterThan(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = ">"

class GreaterEqual(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = ">="

class LessEqual(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = "<="

class In(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = "in"

class Like(Comparison):
    def __init__(self, lhs: Union[str, int, float, Comparison, Statement], rhs: Union[str, int, float, Comparison, Statement]) -> None:
        super().__init__(lhs, rhs)
        self.operator = "in"